#include "pow.h"

void PowCreateGraph(char *graph_name, char *curves, char *images,
		    char *xunits, char *yunits, char *xlabel, char *ylabel,
		    int *xdimdisp, int *ydimdisp, double *xmin_in, 
		    double *ymin_in, double *xmax_in, double *ymax_in, 
		    int *status) {
char whichPowCanvas[9]=".pow.pow";


PowCreateGraph_internal(graph_name, curves, images, xunits, yunits,
			xlabel, ylabel, xdimdisp, ydimdisp, xmin_in, 
			ymin_in, xmax_in, ymax_in, whichPowCanvas, 
			status) ;

}




void PowCreateGraph_internal(char *graph_name, char *curves, char *images,
		    char *xunits, char *yunits, char *xlabel, char *ylabel,
		    int *xdimdisp, int *ydimdisp, double *xmin_in, 
		    double *ymin_in, double *xmax_in, double *ymax_in, 
		    char *whichPowCanvas, int *status) {
  /* xdimdisp and ydimdisp are the *displayed* size of the new graph it will be
     zoomed or shrunk by an integral or 1/integral factor to come as
     close as possible to filling this requested space. If xdimdisp and 
     ydimdisp are both 0, the graph will appear at Magstep 1 */

  PowGraph *graph_instance;
  Tcl_HashEntry *entry_ptr;
  int new = 0;
  double xmin,xmax,ymin,ymax,xdim,ydim,xpixunit,ypixunit;
  int xmargin,ymargin;
  double oldxmag, oldymag, xoff, yoff;
  char bbox[128];
  char extraparams[256];
  char *cptr;
  char *str_ptr;
  char *aspect="no";
  int in_limits;
  int x_points_right,y_points_up;
  
  if (xmin_in != NULL && xmax_in != NULL && *xmin_in > *xmax_in) {
    x_points_right = 0;
  } else {
    x_points_right = 1;
  }

  if (ymin_in != NULL && ymax_in != NULL && *ymin_in > *ymax_in) {
    y_points_up = 0;
  } else {
    y_points_up = 1;
  }


  in_limits = 1;

  entry_ptr = Tcl_CreateHashEntry(&PowGraphTable, graph_name, &new);

#ifdef DEBUG  
  if (!new ) {
    printf("Reusing graph name: %s\n",graph_name);
  }
#endif

  graph_instance = (PowGraph *) ckalloc(sizeof(PowGraph));

  if(graph_instance == NULL) {
    *status = TCL_ERROR;
    fprintf(stderr, "Couldn't ckalloc graph structure space");
    Tcl_DeleteHashEntry(entry_ptr);
    return;
  }

  Tcl_SetHashValue( entry_ptr, graph_instance);

  str_ptr = ckalloc(strlen(graph_name)+1);
  strncpy(str_ptr,graph_name,strlen(graph_name)+1);


  graph_instance->graph_name = str_ptr;

  PowWCSInitGraph( graph_instance, curves, images,
		   x_points_right, y_points_up);
  if( graph_instance->WCS.WCSexists ) aspect="yes";

 
  /* xmin, etc, are initially in graph's "pixel" coordinates */

  xmin =   DBL_MAX;
  xmax = - DBL_MAX;
  ymin =   DBL_MAX;
  ymax = - DBL_MAX;
    
  if( PowFindCurvesBBox(graph_name,curves,&xmin,&xmax,&ymin,&ymax,
			&graph_instance->WCS)
        == TCL_ERROR ) {
     *status =  TCL_ERROR;
     fprintf(stderr, "Error placing curve in graph.\n%s\n", interp->result);
     Tcl_DeleteHashEntry(entry_ptr);
     return;
  };

  if( images != NULL && strstr(images,"NULL") == NULL ) {

     if( PowFindImagesBBox(images,&xmin,&xmax,&ymin,&ymax,
			   &graph_instance->WCS) 
	 == TCL_ERROR ) {
	*status =  TCL_ERROR;
	fprintf(stderr, "Error placing image in graph.\n%s\n", interp->result);
	Tcl_DeleteHashEntry(entry_ptr);
	return;
     }
     aspect = "yes";

  } else if( xmin!=DBL_MAX ) {

     /*  Only plots in graph... expand by 10% for margins around points */
     xdim = xmax - xmin;
     ydim = ymax - ymin;
     xmin -= 0.05*xdim;
     ymin -= 0.05*ydim;
     xmax += 0.05*xdim;
     ymax += 0.05*ydim;

  }

  /*  Convert bounds back into graph coordinates   */

  if( xmin!=DBL_MAX ) {
     PowPixToPos(xmin, ymin, &graph_instance->WCS, &xmin, &ymin );
     PowPixToPos(xmax, ymax, &graph_instance->WCS, &xmax, &ymax );
  } else {
     xmin = xmax = ymin = ymax = 0.0;
  }


  if (xmin_in != NULL) xmin = *xmin_in;
  if (xmax_in != NULL) xmax = *xmax_in;
  if (ymin_in != NULL) ymin = *ymin_in;
  if (ymax_in != NULL) ymax = *ymax_in;

  if (xmin == xmax) {
    if(xmin == 0) {
      xmax = 1;
    } else {
      xmin *= 0.9;
      xmax *= 1.1;
    }
  }

  if (ymin == ymax) {
    if (ymin == 0) {
      ymax = 1;
    } else {
      ymin *= 0.9;
      ymax *= 1.1;
    }
  }

  PowSortGraphMinMax(graph_instance,&xmin,&xmax,&ymin,&ymax,&xdim,&ydim);


  graph_instance->xleft  = xmin;
  graph_instance->xright = xmax;
  graph_instance->ybot   = ymin;
  graph_instance->ytop   = ymax;

  PowPosToPix( xmin, ymin, &graph_instance->WCS, &xoff, &yoff );
  graph_instance->WCS.xrefpix -= xoff;
  graph_instance->WCS.yrefpix -= yoff;

  str_ptr = ckalloc(strlen(xunits)+1);
  strncpy(str_ptr,xunits,strlen(xunits)+1);
  graph_instance->xunits = str_ptr;
  str_ptr = ckalloc(strlen(yunits)+1);
  strncpy(str_ptr,yunits,strlen(yunits)+1);
  graph_instance->yunits = str_ptr;
  str_ptr = ckalloc(strlen(xlabel)+1);
  strncpy(str_ptr,xlabel,strlen(xlabel)+1);
  graph_instance->xlabel = str_ptr;
  str_ptr = ckalloc(strlen(ylabel)+1);
  strncpy(str_ptr,ylabel,strlen(ylabel)+1);
  graph_instance->ylabel = str_ptr;

  
  sprintf(bbox," %#.17g %#.17g %#.17g %#.17g", graph_instance->xleft, 
	  graph_instance->xright, graph_instance->ybot, graph_instance->ytop);

  if( *xdimdisp<=0 ) *xdimdisp = xdim;
  if( *ydimdisp<=0 ) *ydimdisp = ydim;

  xmargin = 80;
  ymargin = 60;

  sprintf(extraparams," %#.17g %#.17g %d %d %s %d %d ",
	  xdim, ydim, *xdimdisp, *ydimdisp, aspect, xmargin, ymargin);


  if ( Tcl_VarEval(interp, "powInitGraph ", graph_name, bbox," {",
		   xunits,"} {", yunits,"} {",xlabel,"} {",ylabel,"} ",
		   whichPowCanvas, extraparams, (char *) NULL) == TCL_ERROR) {
    *status =  TCL_ERROR;
    fprintf(stderr, "Error initializing graph.\n%s\n", interp->result);
    Tcl_DeleteHashEntry(entry_ptr);
    return;
  };

  if( images==NULL ) images="NULL";
  if( curves==NULL ) curves="NULL";
  if ( Tcl_VarEval(interp, "powBuildGraph ", graph_name,
		   " [list ", images," ] ", " [list ", curves," ] ",
		   whichPowCanvas, (char *) NULL)
       == TCL_ERROR) {
     *status =  TCL_ERROR;
     fprintf(stderr, "Error building graph\n%s\n", interp->result);
     Tcl_DeleteHashEntry(entry_ptr);
     return;
  }

  if ( !strcmp( whichPowCanvas, ".pow.pow" ) ) {
     if ( Tcl_VarEval(interp, "powSelectGraph ", graph_name, (char *) NULL)
	  == TCL_ERROR) {
	*status =  TCL_ERROR;
	fprintf(stderr, "Error selecting graph.\n%s\n", interp->result);
	Tcl_DeleteHashEntry(entry_ptr);
	return;
     }
  }
}



void PowDestroyGraph(char *graph_name, int *status) {
  Tcl_HashEntry *entry_ptr;
  char errormsg[1024];
  PowGraph *graph_ptr;

  
  entry_ptr = Tcl_FindHashEntry(&PowGraphTable,graph_name);
  
  if (entry_ptr == NULL) {
    *status = TCL_ERROR;
    sprintf(errormsg,"Can't find POWGraph Object %s to destroy",graph_name);
    Tcl_SetResult(interp,errormsg,TCL_VOLATILE);
    return;
  }

  Tcl_VarEval(interp,"powUnmapGraph ",graph_name,(char *)NULL);

  graph_ptr = Tcl_GetHashValue(entry_ptr);

  /*Delete the entry from the master POWData Hash*/
  Tcl_DeleteHashEntry(entry_ptr);
  
  /*free the PowGraph memory itself and the string holding the name and labels,
     although this is small change*/
  ckfree(graph_ptr->graph_name);
  ckfree(graph_ptr->xunits);
  ckfree(graph_ptr->yunits);
  ckfree(graph_ptr->xlabel);
  ckfree(graph_ptr->ylabel);
  ckfree(graph_ptr);
  return;
}
