###################################################################
#
# TCL routines handling creation and manipulation of SAO regions
#
###################################################################


proc powRegion { } {
    global powPlotParam regionParam powbg currgn currimg
    global powFrameForTop
    global powDWP


    if {[winfo exists ${powDWP}region]} {
	catch {raise ${powDWP}region}
	return
    }
    powToplevel ${powDWP}region .pow "-bg $powbg"
    catch {wm title ${powDWP}region "Edit Region"}
    frame ${powDWP}region.head -bg $powbg
    label ${powDWP}region.head.title -bg yellow -fg black
    button ${powDWP}region.head.help -text "Help" \
	    -command {powHelp Regions.html} \
	    -bg $powbg
    pack ${powDWP}region.head.title -in ${powDWP}region.head -side left
    pack ${powDWP}region.head.help -in ${powDWP}region.head -side right -padx 2 -pady 2

    frame ${powDWP}region.list -bg $powbg
    scrollbar ${powDWP}region.list.scrolly -orient vertical \
	    -command {${powDWP}region.list.rgns yview} -bg $powbg
    scrollbar ${powDWP}region.list.scrollx -orient horizontal \
	    -command {${powDWP}region.list.rgns xview} -bg $powbg
    listbox ${powDWP}region.list.rgns -width 30 -height 6 \
	    -selectmode browse -bg $powbg -exportselection 0 \
	    -yscrollcommand "${powDWP}region.list.scrolly set " \
	    -xscrollcommand "${powDWP}region.list.scrollx set "
    bind ${powDWP}region.list.rgns <ButtonRelease-1> \
	    {powSelectRegion [${powDWP}region.list.rgns curselection]}

    grid ${powDWP}region.list.rgns    -in ${powDWP}region.list -row 1 -column 1 \
	    -sticky news
    grid ${powDWP}region.list.scrolly -in ${powDWP}region.list -row 1 -column 2 \
	    -sticky news
    grid ${powDWP}region.list.scrollx -in ${powDWP}region.list -row 2 -column 1 \
	    -sticky news
    grid rowconfigure ${powDWP}region.list 1 -weight 1
    grid columnconfigure ${powDWP}region.list 1 -weight 1

    frame ${powDWP}region.currshape1 -bg $powbg
    label ${powDWP}region.currshape1.lbl -text "Current Shape: " -bg $powbg
    button ${powDWP}region.currshape1.apply -text "Apply" -bg $powbg \
	    -command {powChangeRegion}
    button ${powDWP}region.currshape1.delete -text "Delete" -bg $powbg \
	    -command {powDeleteCurrRegion}

    pack ${powDWP}region.currshape1.lbl -in ${powDWP}region.currshape1 \
	    -side left
    pack ${powDWP}region.currshape1.apply -in ${powDWP}region.currshape1 \
	    -side right -padx 5
    pack ${powDWP}region.currshape1.delete -in ${powDWP}region.currshape1 \
	    -side right -padx 5

    frame ${powDWP}region.currshape2 -bg $powbg

#    if $powFrameForTop {
#	set regionParam(supportedShapes) Polygon
#    } else {
	menu ${powDWP}region.shapemenu -tearoff 0 -bg $powbg
	set regionParam(supportedShapes) \
	    [list Box Circle Ellipse Polygon Line Point]
	foreach s $regionParam(supportedShapes) {
	    ${powDWP}region.shapemenu add command -label $s \
		-command "powChangeShape $s"
	}
#    }

    button ${powDWP}region.currshape2.shapebut \
	    -textvariable regionParam(currShape) \
	    -bg $powbg -relief raised -width 8
    bind ${powDWP}region.currshape2.shapebut <ButtonPress-1> {
	tk_popup ${powDWP}region.shapemenu \
		[winfo pointerx ${powDWP}region.shapemenu] \
		[winfo pointery ${powDWP}region.shapemenu] \
		[lsearch -exact $regionParam(supportedShapes) \
		                $regionParam(currShape) ]
	${powDWP}region.currshape2.shapebut configure -relief raised
    }
    entry ${powDWP}region.currshape2.desc -width 35

    button ${powDWP}region.currshape2.signbut \
	    -textvariable regionParam(currSign) \
	    -bg $powbg -relief raised -width 1 \
	    -command {
	if {$regionParam(currSign)=="+"} \
		{set regionParam(currSign) "-" } \
		else   {set regionParam(currSign) "+"}
    }

    pack ${powDWP}region.currshape2.signbut -in ${powDWP}region.currshape2 \
	    -side left -padx 1
    pack ${powDWP}region.currshape2.shapebut -in ${powDWP}region.currshape2 \
	    -side left -padx 1
    pack ${powDWP}region.currshape2.desc -in ${powDWP}region.currshape2 \
	    -side left -padx 1 -fill x -expand 1

    frame ${powDWP}region.radiobtns -bg $powbg
    radiobutton ${powDWP}region.radiobtns.deg -text Degrees -value degrees \
	    -variable regionParam(coords) -bg $powbg -highlightthickness 0 \
	    -command { powUpdateRegionList }
    radiobutton ${powDWP}region.radiobtns.pix -text Pixels -value pixels \
	    -variable regionParam(coords) -bg $powbg -highlightthickness 0 \
	    -command { powUpdateRegionList }
    pack ${powDWP}region.radiobtns.deg -in ${powDWP}region.radiobtns -side left -padx 10
    pack ${powDWP}region.radiobtns.pix -in ${powDWP}region.radiobtns -side left -padx 10

    frame ${powDWP}region.btns -bg $powbg
    button ${powDWP}region.btns.exit -text "Exit" -command powExitRegionDlg \
	    -bg $powbg
    button ${powDWP}region.btns.clear -text "Clear All" -command powClearRegions \
	    -bg $powbg
    button ${powDWP}region.btns.save -text "Save..." -command powSaveRegionFile \
	    -bg $powbg
    button ${powDWP}region.btns.open -text "Open..." -command powOpenRegionFile \
	    -bg $powbg
    pack ${powDWP}region.btns.open -in ${powDWP}region.btns -side left -padx 7
    pack ${powDWP}region.btns.save -in ${powDWP}region.btns -side left -padx 7
    pack ${powDWP}region.btns.clear -in ${powDWP}region.btns -side left -padx 7
    pack ${powDWP}region.btns.exit -in ${powDWP}region.btns -side left -padx 7

    set bbox [.pow.pow coords ${currgn}box]
    foreach {x1 y1 x2 y2} $bbox {}
    set boxid [.pow.pow create polygon $x1 $y1 $x1 $y2 $x2 $y2 $x2 $y1 \
	    -fill {} -tags "$currgn DrawRegion"]

    powBindBtn <<RGN_Create>> ".pow.pow bind DrawRegion" \
	  { powCreateNewShape %x %y } \
	  { powDragNewShape   %x %y } \
	  { powFinishNewShape -1 }
	   
    .pow.pow bind DragAble <Enter> {.pow.pow configure -cursor fleur}
    .pow.pow bind DragAble <Leave> {.pow.pow configure -cursor $powcursor}

    .pow.pow bind rgnHandle <Enter> {.pow.pow configure -cursor sizing}
    .pow.pow bind rgnHandle <Leave> {.pow.pow configure -cursor $powcursor}

    bind .pow <Shift-KeyPress-Up>    { powShiftRegion1 0 -1 }
    bind .pow <Shift-KeyPress-Left>  { powShiftRegion1 -1 0 }
    bind .pow <Shift-KeyPress-Right> { powShiftRegion1 1 0 }
    bind .pow <Shift-KeyPress-Down>  { powShiftRegion1 0 1 }

    bind .pow <Control-KeyPress-Up>    { powShiftRegion1   0 -10 }
    bind .pow <Control-KeyPress-Left>  { powShiftRegion1 -10   0 }
    bind .pow <Control-KeyPress-Right> { powShiftRegion1  10   0 }
    bind .pow <Control-KeyPress-Down>  { powShiftRegion1   0  10 }

    grid ${powDWP}region.head -in ${powDWP}region -column 1 -row 0 -sticky ew
    grid ${powDWP}region.list -in ${powDWP}region -column 1 -row 1 \
	    -sticky news
    grid ${powDWP}region.currshape1 -in ${powDWP}region -column 1 -row 3 \
	    -sticky ew
    grid ${powDWP}region.currshape2 -in ${powDWP}region -column 1 -row 4 \
	    -sticky ew
    grid ${powDWP}region.radiobtns -in ${powDWP}region -column 1 -row 2 -pady 5
    grid ${powDWP}region.btns -in ${powDWP}region -column 1 -row 6 -pady 5

    grid columnconfigure ${powDWP}region 1 -weight 1
    grid rowconfigure ${powDWP}region 1 -weight 1
    grid rowconfigure ${powDWP}region 2 -minsize 5
    grid rowconfigure ${powDWP}region 5 -minsize 10
    catch {wm minsize ${powDWP}region 300 200}

    set regionParam(canvasID) $boxid
    if { $regionParam(nItems)==0 } {
	set regionParam(gn)       $currgn
	set regionParam(allIDs)   ""
	set regionParam(allSigns) ""
	set regionParam(allShapes) ""
	set regionParam(allDescr) ""
	set regionParam(currID)   ""
	set regionParam(currSign) "+"
	set regionParam(currShape) "Circle"
	set regionParam(currDescr) "0 0 0 0 0"

	if {[powWCSexists $currgn]} {
	    set fmt "degrees"
	} else {
	    set fmt "pixels"
	    ${powDWP}region.radiobtns.deg configure -state disabled
	}
	set regionParam(coords)   $fmt
    }

    bind ${powDWP}region.list.rgns <Destroy> {powExitRegionDlg}
    powRedrawRegions
    powUpdateRegionTitle
    powUpdateRegionList
}

proc powExitRegionDlg { } {
    global regionParam
    global powDWP

    destroy ${powDWP}region
    catch {.pow.pow delete $regionParam(canvasID)}
    catch {.pow.pow dtag DrawRegion DrawRegion}
}

proc powUpdateRegionTitle { } {
    global regionParam
    global powDWP

    if { [info exists regionParam(regfile)] && $regionParam(regfile)!=""} {
	${powDWP}region.head.title configure \
		-text "Regions for $regionParam(gn) ($regionParam(regfile))"
    } else {
	${powDWP}region.head.title configure -text "Regions for $regionParam(gn)"
    }
}

proc powSaveRegionFile { } {
    global regionParam currimg

    if { [info exists regionParam(regfile)] && $regionParam(regfile)!="" } {
	set defFile $regionParam(regfile)
    } else {
	set defFile pow.reg
    }
    set filename [tk_getSaveFile -initialfile "$defFile"]
    if {$filename ==  "" } return
    set i [expr [string last "/" $filename] + 1 ]
    if {$i<1} {set i 0}
    set regionParam(regfile) [string range $filename $i end]
    powUpdateRegionTitle

    set freg [open $filename w]
    set fmt $regionParam(coords)
    puts $freg "# Region created by POW [clock format [clock seconds]]"
    puts $freg "# filename: $regionParam(gn)"
    puts $freg "# format: $fmt"
    for {set i 0} {$i<$regionParam(nItems)} {incr i} {
	set sign  [lindex $regionParam(allSigns) $i]
	set shape [lindex $regionParam(allShapes) $i]
	set descr [powConvDesc2Fctn $shape \
		[lindex $regionParam(allDescr) $i] $fmt ]
	puts $freg "$sign[string tolower $shape]$descr"
    }
    close $freg
}

proc powOpenRegionFile { } {
    global regionParam
    global powDWP

    set types {
	{{Region Files}       {.reg}        }
         {{All Files}        *             }
    }

    if { [info exists regionParam(regfile)] && $regionParam(regfile)!="" } {
	set defFile $regionParam(regfile)
    } else {
	set defFile ""
    }
    set filename [tk_getOpenFile -filetypes $types -initialfile $defFile]
    if {$filename ==  "" } return
    set freg [open $filename r]
    set stuff [read $freg]
    close $freg
    
    if { $regionParam(nItems) } {
	set act [tk_dialog ${powDWP}regionInquiry "Open Region File" \
		"Region files already exist" warning 2 Cancel Overwrite Append]
	if { $act==-1 || $act==0 } {return}
	if { $act==1 } {
	    while { $regionParam(nItems) != 0 } { powDeleteCurrRegion }
	}
    }

    set i [expr [string last "/" $filename] + 1 ]
    if {$i<1} {set i 0}
    set regionParam(regfile) [string range $filename $i end]
    powUpdateRegionTitle

    set newregions [split $stuff "\n"]
    set fmt "default"
    foreach reg $newregions {
	if {[string index $reg 0] == "#"} {
	    set reg [string tolower $reg]
	    if { [string first "format:" $reg]!=-1 } {
		if { [string first "degrees" $reg]!=-1 } {
		    set fmt "degrees"
		} elseif { [string first "pixels" $reg]!=-1 } {
		    set fmt "pixels"
		}
	    }
	} elseif {[string length $reg] != 0} {
	    if { [powAddNewRegion $reg $fmt] } {
		puts "Could not decipher $reg"
		break
	    }
	}
    }
}

		  
proc powRedrawRegions { } {
    global currgn regionParam
    global powDWP

    if { $regionParam(gn)!=$currgn } {return}

    if { [.pow.pow find withtag $regionParam(canvasID)] == "" \
	    && [winfo exists ${powDWP}region] } {
	set bbox [.pow.pow coords ${currgn}box]
	foreach {x1 y1 x2 y2} $bbox {}
	set regionParam(canvasID) [.pow.pow create polygon \
		$x1 $y1 $x1 $y2 $x2 $y2 $x2 $y1 \
		-fill {} -tags "$currgn DrawRegion"]
    }

    if { $regionParam(nItems)==0 } return

    set ids ""
    for {set i 0} {$i<$regionParam(nItems)} {incr i} {
	set sign  [lindex $regionParam(allSigns) $i]
	set shape [lindex $regionParam(allShapes) $i]
	set descr [lindex $regionParam(allDescr) $i]
	set id    [lindex $regionParam(allIDs) $i]

	.pow.pow delete [lindex $regionParam(allIDs) $i]
	set newID [powDrawShape $sign $shape $descr 0]
	powBindRegion $newID
	lappend ids $newID

	if { $id == $regionParam(currID) } {
	    set regionParam(currID) $newID
	}
    }
    set regionParam(allIDs) $ids
    powCreateRegionHandles
}

proc powDrawShape { sign shape descr {new 0} } {
    global regionParam

    if { ! $new } { set descr [powConv2NewDescr $shape $descr] }

# descr now in standard, internal, format--rot x0 y0 x1 y1...--in canvas coords

    if {$sign=="+"} {set color lightblue} else {set color red}
    set coords [ShapeToPoly $shape $descr]

    if {$new} {
	set item [eval [concat .pow.pow create polygon \
		$coords -outline $color -fill {{}} -tags DragAble] ]
	set rgnID "rgnID$item"
	.pow.pow addtag $rgnID withtag $item
    } else {
	set bbox [.pow.pow coords $regionParam(gn)box]
	foreach {x1 y1 x2 y2} $bbox {}
	set bbox "$x1 $y1 $x1 $y2 $x2 $y2 $x2 $y1 $x1 $y1"
	set coords [powClipPolys $coords $bbox]
	if {$coords==""} { return "offscreen" }
	foreach coords1 $coords {
	    set item [eval [concat .pow.pow create polygon \
		    $coords1 -outline $color -fill {{}} -tags DragAble] ]
	    if { ![info exists rgnID] } {
		set rgnID "rgnID$item"
	    }
	    .pow.pow addtag $rgnID withtag $item
	}
    }
    return $rgnID
}

proc powCreateNewShape { x y } {
    global regionParam

    set x [.pow.pow canvasx $x]
    set y [.pow.pow canvasy $y]

    set regionParam(currID) ""
    set regionParam(currDescr) "0 $x $y $x $y"
    set regionParam(currID) [powDrawShape $regionParam(currSign) \
	    $regionParam(currShape) $regionParam(currDescr) 1]
}

proc powDragNewShape { x y } {
    global regionParam

    set x [.pow.pow canvasx $x]
    set y [.pow.pow canvasy $y]

    if {$regionParam(currShape)=="Point"} {
	set regionParam(currDescr) \
		[lreplace $regionParam(currDescr) 1 2 $x $y]
    } elseif {$regionParam(currShape)=="Box"} {
	foreach {x0 y0 x2 y2} [lrange $regionParam(currDescr) 1 4] {}
	set x1 [expr $x0-$x2+$x0]
	set y1 [expr $y0-$y2+$y0]
	set x0 [expr 0.5*($x1+$x)]
	set y0 [expr 0.5*($y1+$y)]
	set regionParam(currDescr) \
		[lreplace $regionParam(currDescr) 1 4 $x0 $y0 $x $y]
    } else {
	set regionParam(currDescr) \
		[lreplace $regionParam(currDescr) 3 4 $x $y]
    }

    .pow.pow delete $regionParam(currID)
    set regionParam(currID) [powDrawShape $regionParam(currSign) \
	    $regionParam(currShape) $regionParam(currDescr) 1]
}

proc powBindRegion { id } {
    global regionParam

    .pow.pow bind $id <Enter> \
	    ".pow.pow itemconfig $id -width 2"
    .pow.pow bind $id <Leave> \
	    ".pow.pow itemconfig $id -width 1"

    powBindBtn <<RGN_Drag>> ".pow.pow bind $id" \
	    "powSelectThisRegion $id ; powBeginDragCurr %x %y" \
	    { powDragCurrShape %x %y } \
	    { powFinishCurrShape }

    # This binding prevents RGN_Drag from executing if RGN_Create is
    # more appropriate due to modifiers
    powBindBtn <<RGN_Create>> ".pow.pow bind $id" {} {} {}

    .pow.pow bind $id <Double-Button-1> \
	    { powRegion }

    .pow.pow addtag $regionParam(gn) withtag $id

# DrawRegion tag allows use of Button 3 to draw new shapes over this object
    .pow.pow addtag DrawRegion withtag $id

}

proc powFinishNewShape { itemNo } {
    global regionParam

    if {$itemNo >= $regionParam(nItems) || $itemNo<0 } {
	if { $itemNo<0 } {
	    set regionParam(currDescr) [powConvNewDescr \
		    $regionParam(currShape) $regionParam(currDescr)]
	}
	.pow.pow delete $regionParam(currID)
	set regionParam(currID) [powDrawShape $regionParam(currSign) \
		$regionParam(currShape) $regionParam(currDescr)]
	lappend regionParam(allIDs)    $regionParam(currID)
	lappend regionParam(allSigns)  $regionParam(currSign)
	lappend regionParam(allShapes) $regionParam(currShape)
	lappend regionParam(allDescr)  $regionParam(currDescr)
	set itemNo $regionParam(nItems)
	incr regionParam(nItems)
    } else {
	.pow.pow delete $regionParam(currID)
	set regionParam(currID) [powDrawShape $regionParam(currSign) \
		$regionParam(currShape) $regionParam(currDescr)]
	set regionParam(allIDs) [lreplace $regionParam(allIDs) $itemNo \
		$itemNo $regionParam(currID)]
	set regionParam(allSigns) [lreplace $regionParam(allSigns) $itemNo \
		$itemNo $regionParam(currSign)]
	set regionParam(allShapes) [lreplace $regionParam(allShapes) $itemNo \
		$itemNo $regionParam(currShape)]
	set regionParam(allDescr) [lreplace $regionParam(allDescr) $itemNo \
		$itemNo $regionParam(currDescr)]
    }

    powBindRegion $regionParam(currID)
    powUpdateRegionList
    powSelectRegion $itemNo
}

proc powBeginDragCurr { x y } {
    global lastX lastY newRgnID

    set x [.pow.pow canvasx $x]
    set y [.pow.pow canvasy $y]

    set lastX $x
    set lastY $y
    set newRgnID -1
}

proc powDragCurrShape { x y } {
    global regionParam lastX lastY newRgnID powcursor powRgnMoveCallback

    set x [.pow.pow canvasx $x]
    set y [.pow.pow canvasy $y]

    set dx [expr $x-$lastX ]
    set dy [expr $y-$lastY ]
    set lastX $x
    set lastY $y

    if {$newRgnID==$regionParam(currID)} {return}
    if {$newRgnID>0} {
	.pow.pow delete $newRgnID
    } else {
	set regionParam(currDescr) \
		[powConv2NewDescr $regionParam(currShape) \
		$regionParam(currDescr)]
	.pow.pow itemconfigure $regionParam(currID) -outline {}
    }

    set descr $regionParam(currDescr)
    set newDescr [lindex $descr 0]
    foreach {x y} [lrange $descr 1 end] {
	lappend newDescr [expr $x+$dx] [expr $y+$dy]
    }
    set regionParam(currDescr) $newDescr

    set newRgnID [powDrawShape $regionParam(currSign) \
	    $regionParam(currShape) $regionParam(currDescr) 1]
    .pow.pow itemconfigure rgnHandle -outline {}
#    .pow.pow move rgnHandle $dx $dy
    if [info exists powRgnMoveCallback] {
	$powRgnMoveCallback $dx $dy
    }
    .pow.pow move $regionParam(currID) $dx $dy
    .pow.pow raise $regionParam(currID)
}

proc powFinishCurrShape {  } {
    global regionParam newRgnID

    if {$newRgnID < 0 || $newRgnID==$regionParam(currID)} {return}

    set regionParam(currDescr) \
	    [powConvNewDescr $regionParam(currShape) $regionParam(currDescr)]

    set currItm [powFindThisRegion $regionParam(currID)]
    .pow.pow delete $regionParam(currID)
    set regionParam(currID) $newRgnID

    powFinishNewShape $currItm
}

proc powInsertDragPt { ptNum } {
    global regionParam

    set descr [powConv2NewDescr $regionParam(currShape) \
	    $regionParam(currDescr)]
    if { $regionParam(currShape) == "Polygon" } {
	set start [expr $ptNum*2+1]
	foreach {x y} [lrange $descr $start [expr $start+1] ] {}
	set descr [linsert $descr $start $x $y]
    }
    set regionParam(currDescr) $descr
}

proc powDragPt { ptNum x y } {
    global regionParam

    set x [.pow.pow canvasx $x]
    set y [.pow.pow canvasy $y]

    set descr $regionParam(currDescr)

    set start [expr $ptNum*2+1]
    set end [expr $start+1]

    set rot [expr -[lindex $descr 0]]
    if { abs($rot)>1e-10 } {
	set x0 [lindex $descr 1]
	set y0 [lindex $descr 2] 
	foreach {x y} [poly_rotate $x0 $y0 $rot [list $x $y]] {}
    }
    set regionParam(currDescr) [lreplace $descr $start $end $x $y]

    set currItm [powFindThisRegion $regionParam(currID)]
    .pow.pow delete $regionParam(currID)
    set regionParam(currID) [powDrawShape $regionParam(currSign) \
	    $regionParam(currShape) $regionParam(currDescr) 1]
    set regionParam(allIDs) [lreplace $regionParam(allIDs) \
	    $currItm $currItm $regionParam(currID)]
}

proc powFinishDragPt { ptNum } {
    global regionParam

    set descr $regionParam(currDescr)
    if {$regionParam(currShape)=="Polygon"} {
	set npts [llength $descr]
	set start1 [expr $ptNum*2+1]
	foreach {x1 y1} [lrange $descr $start1 [expr $start1+1] ] {}

	set start0 [expr $start1-2]
	if {$start0<1} {set start0 [expr $npts-2]}
	foreach {x0 y0} [lrange $descr $start0 [expr $start0+1] ] {}
	set dx [expr $x1-$x0]
	set dy [expr $y1-$y0]
	if { [expr $dx*$dx+$dy*$dy]<16 } {
	    set descr [lreplace $descr $start1 [expr $start1+1] ]
	} else {

	    set start0 [expr $start1+2]
	    if {$start0>=$npts} {set start0 1}
	    foreach {x0 y0} [lrange $descr $start0 [expr $start0+1] ] {}
	    set dx [expr $x1-$x0]
	    set dy [expr $y1-$y0]
	    if { [expr $dx*$dx+$dy*$dy]<16 } {
		set descr [lreplace $descr $start1 [expr $start1+1] ]
	    }
	}
    } 

    set regionParam(currDescr) \
	    [powConvNewDescr $regionParam(currShape) $descr]

    powFinishNewShape [powFindThisRegion $regionParam(currID)]
}

proc powRotatePt { x y } {
    global regionParam

    set x [.pow.pow canvasx $x]
    set y [.pow.pow canvasy $y]

    set descr $regionParam(currDescr)
    set dx   [expr $x-[lindex $descr 1]]
    set dy   [expr $y-[lindex $descr 2]]

    set rot [expr -atan2($dy,$dx)*180/3.1415962]

    set regionParam(currDescr) [lreplace $descr 0 0 $rot]

    set currItm [powFindThisRegion $regionParam(currID)]
    .pow.pow delete $regionParam(currID)
    set regionParam(currID) [powDrawShape $regionParam(currSign) \
	    $regionParam(currShape) $regionParam(currDescr) 1]
    set regionParam(allIDs) [lreplace $regionParam(allIDs) \
	    $currItm $currItm $regionParam(currID)]
    .pow.pow itemconfigure rgnHandle -outline {}
}

proc powFinishRotatePt {  } {
    global regionParam

    set regionParam(currDescr) \
	    [powConvNewDescr $regionParam(currShape) $regionParam(currDescr)]

    powFinishNewShape [powFindThisRegion $regionParam(currID)]
}

proc powUpdateRegionList { } {
    global regionParam
    global powDWP

    if { ![winfo exists ${powDWP}region] } { return }

    set nItems $regionParam(nItems)
    set allSigns $regionParam(allSigns)
    set allShapes $regionParam(allShapes)
    set allDescr $regionParam(allDescr)

    set currItm [${powDWP}region.list.rgns curselection]
    ${powDWP}region.list.rgns delete 0 end
    set n 0
    while {$n<$nItems} {
	set sign [lindex $allSigns $n]
	set shape [lindex $allShapes $n]
	set descr [lindex $allDescr $n]
	set descr [powConvDesc2Fctn $shape $descr]
	set txtDescr "$sign${shape}$descr"
	${powDWP}region.list.rgns insert end $txtDescr
	if {$currItm!="" && $n==$currItm} {
	    ${powDWP}region.currshape2.desc delete 0 end
	    ${powDWP}region.currshape2.desc insert 0 $descr
	}
	incr n
    }
    if {$currItm!=""} {
	${powDWP}region.list.rgns selection set $currItm
	${powDWP}region.list.rgns see $currItm
    }
    if { $currItm=="" && $nItems>0 } {
	powSelectThisRegion $regionParam(currID)
    }
}

proc powSelectThisRegion { itemID } {
    powSelectRegion [powFindThisRegion $itemID]
}

proc powFindThisRegion { itemID } {
    global regionParam

    set nItems $regionParam(nItems)
    for {set i 0} {$i<$nItems} {incr i} {
	if { $itemID == [lindex $regionParam(allIDs) $i] } {
	    return $i
	}
    }
    puts "$itemNo not found!"
}

proc powSelectRegion { itemNo } {
    global regionParam
    global powDWP

    if {$itemNo==""} {return}
    set item  [lindex $regionParam(allIDs) $itemNo]
    set shape [lindex $regionParam(allShapes) $itemNo]
    set sign  [lindex $regionParam(allSigns) $itemNo]
    set descr [lindex $regionParam(allDescr) $itemNo]

    set regionParam(currShape) $shape
    set regionParam(currSign) $sign
    set regionParam(currDescr) $descr
    set regionParam(currID) $item

    .pow.pow raise $item
    powCreateRegionHandles

    if { [winfo exists ${powDWP}region] } {
	${powDWP}region.currshape2.desc delete 0 end
	${powDWP}region.currshape2.desc insert 0 [powConvDesc2Fctn $shape $descr]
	${powDWP}region.list.rgns select clear 0 end
	${powDWP}region.list.rgns select set $itemNo
	${powDWP}region.list.rgns see $itemNo
    }
}

proc powShiftRegion1 { dx dy } {
    global regionParam

    if {$regionParam(nItems)==0} {return}

    set descr [powConv2NewDescr $regionParam(currShape) \
		$regionParam(currDescr)]

    set newDescr [lindex $descr 0]
    foreach {x y} [lrange $descr 1 end] {
	lappend newDescr [expr $x+$dx] [expr $y+$dy]
    }
    set regionParam(currDescr) \
	    [powConvNewDescr $regionParam(currShape) $newDescr]
    powFinishNewShape [powFindThisRegion $regionParam(currID)]

    update idletasks
}

proc powCreateRegionHandles { } {
    global regionParam

    .pow.pow delete rgnHandle

    set shape $regionParam(currShape)
    if {$shape=="Point"} {return}
    set newDescr [powConv2NewDescr $shape $regionParam(currDescr)]
    set gn $regionParam(gn)

    if { $shape=="Line" || $shape=="Polygon" } {
	set start 1
	set end   [llength $newDescr]
    } else {
	set start 3
	set end   4
    }
    set rot [lindex $newDescr 0]
    set x0 [lindex $newDescr 1]
    set y0 [lindex $newDescr 2]
    foreach {bx1 by1 bx2 by2} [.pow.pow coords ${gn}box] {}

# Create Move Point Handles

    set ptNum [expr ($start-1)/2]
    foreach {x y} [lrange $newDescr $start $end] {
	foreach {x y} [poly_rotate $x0 $y0 $rot [list $x $y] ] {}
	if { $x<$bx2 && $x>$bx1 && $y<$by2 && $y>$by1 } {
	    set x1 [expr $x-2]
	    set y1 [expr $y-2]
	    set x2 [expr $x+2]
	    set y2 [expr $y+2]
	    set id [.pow.pow create polygon $x1 $y1 $x1 $y2 $x2 $y2 $x2 $y1 \
		    -outline green -fill {} -tags "rgnMovePt rgnHandle $gn"]

	    powBindBtn <<RGN_DragPt>> ".pow.pow bind $id" \
		  { set regionParam(currDescr) \
                         [powConv2NewDescr $regionParam(currShape) \
		         $regionParam(currDescr)] } \
		  "powDragPt $ptNum %x %y" \
		  "powFinishDragPt $ptNum"

	    powBindBtn <<RGN_InsertPt>> ".pow.pow bind $id" \
		  "powInsertDragPt $ptNum" \
		  "powDragPt $ptNum %x %y" \
		  "powFinishDragPt $ptNum"
	}
	incr ptNum
    }

# Create Rotate Region Handle... only for Box and Ellipse

    if {$shape=="Box" || $shape=="Ellipse"} {
	set dx [expr [lindex $newDescr 3]-$x0]
	if { $shape=="Ellipse" } {set dx [expr 1.41421356*$dx]}
	set x [expr $x0+$dx]
	set y [lindex $newDescr 2]
	foreach {x y} [poly_rotate $x0 $y0 $rot [list $x $y] ] {}
	if { $x<$bx2 && $x>$bx1 && $y<$by2 && $y>$by1 } {
	    set x1 [expr $x-2]
	    set y1 [expr $y-2]
	    set x2 [expr $x+2]
	    set y2 [expr $y+2]
	    set id [.pow.pow create polygon $x1 $y2 $x2 $y2 $x $y1 \
		    -outline green -fill {} -tags "rgnRotate rgnHandle $gn"]

	    powBindBtn <<RGN_Rotate>> ".pow.pow bind $id" \
		  { set regionParam(currDescr) \
		        [powConv2NewDescr $regionParam(currShape) \
		                          $regionParam(currDescr)] } \
		  "powRotatePt %x %y" \
		  "powFinishRotatePt"
	}
    }
}

proc powChangeShape { newShape } {
    global regionParam
    global powDWP

    set Descr [powConvFctn2Desc [${powDWP}region.currshape2.desc get] ]
    if {$Descr == ""} {
	set regionParam(currShape) $newShape
	${powDWP}region.currshape2.desc delete 0 end
	return
    }

    foreach {oldSign oldShape oldDescr} $Descr {}
    if {$oldShape==$newShape} {return}

    set stdDescr [powConv2NewDescr $oldShape $oldDescr]
    if {$oldShape=="Polygon"} {
	set sumX 0
	set sumY 0
	set sumXX 0
	set sumYY 0
	set sumXY 0
	set cnt 0
	foreach {x y} [lrange $stdDescr 1 end] {
	    set sumX [expr $sumX+$x]
	    set sumY [expr $sumY+$y]
	    set sumXX [expr $sumXX+$x*$x]
	    set sumYY [expr $sumYY+$y*$y]
	    set sumXY [expr $sumXY+$x*$y]
	    incr cnt
	}
	set x1 [expr $sumX/$cnt] 
	set y1 [expr $sumY/$cnt]
	set dx [expr sqrt($sumXX/$cnt-$x1*$x1)]
	set dy [expr sqrt($sumYY/$cnt-$y1*$y1)]
	set theta 0
# 	puts "\n$x1 $y1 $dx $dy $theta"
# 	if {$newShape!="Line"} {
# 	    set sumDxDy [expr $sumXY-$cnt*$x1*$y1]
# 	    set sumDxDx [expr $sumXX-$cnt*$x1*$x1]
# 	    set theta [expr -atan($sumDxDy/$sumDxDx)]
# 	    if {$theta<-0.79} {set theta [expr $theta+3.1415962/2.0]}
# 	    set Dx [expr  $dx*cos($theta)+$dy*sin($theta)]
# 	    set Dy [expr -$dx*sin($theta)+$dy*cos($theta)]
# 	    foreach {dx dy} "$Dx $Dy" {}
# 	    set theta [expr $theta * 180.0/3.1415962]
# 	}
# 	puts "$x1 $y1 $dx $dy $theta"
	set x2 [expr $x1+$dx]
	set y2 [expr $y1+$dy]
	set stdDescr "$theta $x1 $y1 $x2 $y2"
    }
    if {$newShape=="Polygon"} {
	set pts [ShapeToPoly $oldShape $stdDescr]
	set npts [expr [llength $pts]-3]
	set stdDescr "0 [lrange $pts 0 $npts]"
    }
    set newDescr [powConvNewDescr $newShape $stdDescr]
    ${powDWP}region.currshape2.desc delete 0 end
    ${powDWP}region.currshape2.desc insert 0 [powConvDesc2Fctn $newShape $newDescr]
    set regionParam(currShape) $newShape
}

proc powChangeRegion { } {
    global regionParam
    global powDWP

    set itemNo [${powDWP}region.list.rgns curselection]
    if {$itemNo==""} {set itemNo $regionParam(nItems)}
    set newDescr [powConvFctn2Desc [${powDWP}region.currshape2.desc get] ]
    if {$newDescr == ""} {return}
    foreach {sign shape descr} $newDescr {}

    .pow.pow delete [lindex $regionParam(allIDs) $itemNo]
    set id [powDrawShape $sign $shape $descr]

    set regionParam(currID)    $id
    set regionParam(currSign)  $sign
    set regionParam(currShape) $shape
    set regionParam(currDescr) $descr

    powFinishNewShape $itemNo
}

proc powAddNewRegion { descr format } {
    global regionParam currimg

    set itemNo $regionParam(nItems)
    set newDescr [powConvFctn2Desc $descr $format]
    if {$newDescr == ""} {return 1}

    foreach {sign shape descr} $newDescr {}
    set id [powDrawShape $sign $shape $descr]

    set regionParam(currID)    $id
    set regionParam(currSign)  $sign
    set regionParam(currShape) $shape
    set regionParam(currDescr) $descr

    powFinishNewShape $itemNo
    return 0
}

proc powDeleteCurrRegion { } {
    global regionParam
    global powDWP

#returns deleted item number for possible use by calling routine    

    if { [winfo exists ${powDWP}region.list.rgns] } {
	set itemNo [${powDWP}region.list.rgns curselect]
    } else {set itemNo 0}
    if {$itemNo!=""} {
	.pow.pow delete [lindex $regionParam(allIDs) $itemNo]
	${powDWP}region.list.rgns delete $itemNo
	set regionParam(allIDs) [lreplace $regionParam(allIDs) \
		$itemNo $itemNo]
	set regionParam(allSigns) [lreplace $regionParam(allSigns) \
		$itemNo $itemNo]
	set regionParam(allShapes) [lreplace $regionParam(allShapes) \
		$itemNo $itemNo]
	set regionParam(allDescr) [lreplace $regionParam(allDescr) \
		$itemNo $itemNo]
	incr regionParam(nItems) -1
	if {$regionParam(nItems)>0} {
	    if {$itemNo==$regionParam(nItems)} {
		powSelectRegion [expr $itemNo-1]
	    } else {
		powSelectRegion $itemNo
	    }
	} else {
	    .pow.pow delete rgnHandle
	}
    }
    return $itemNo
}

proc powClearRegions { } {
    global regionParam powDWP

    set act [tk_messageBox -message "Delete All regions?" -type yesno \
	    -default no]

    if { $act=="yes" } {
	foreach id $regionParam(allIDs) { .pow.pow delete $id }
	.pow.pow delete rgnHandle
	${powDWP}region.list.rgns delete 0 end
	set regionParam(allIDs)   ""
	set regionParam(allSigns) ""
	set regionParam(allShapes) ""
	set regionParam(allDescr) ""
	set regionParam(nItems) 0
    }
}

#########
#
#   Convert descriptions between various formats
#

proc powConvNewDescr { shape descr } {
# Convert standard 2n+1 parameters to the shape specific description,
# where n is the number of points.  First parameter is rotation.

    global regionParam powPlotParam
    
    set cnt 0
    set rot [lindex $descr 0]
    set gn $regionParam(gn)
    foreach {x y} [lrange $descr 1 end] {
	set x$cnt $x
	set y$cnt $y
	foreach "X$cnt Y$cnt" [powCanvasToGraph $gn $x $y] {}
	incr cnt
    }
    set dx [expr abs($x1-$x0)/$powPlotParam(xmagstep,$gn)]
    set dy [expr abs($y1-$y0)/$powPlotParam(ymagstep,$gn)]
    
    switch $shape {
	Box     { set descr "$X0 $Y0 [expr $dx+$dx] [expr $dy+$dy] $rot" }
	Circle  { set descr "$X0 $Y0 [expr sqrt($dx*$dx+$dy*$dy)]" }
	Ellipse { 
	    set descr "$X0 $Y0 [expr 1.41421356*$dx] [expr 1.41421356*$dy] $rot"
	}
        Polygon {
	    set descr "$X0 $Y0"
	    for {set i 1} {$i<$cnt} {incr i} {
		eval [concat lappend descr \$X$i \$Y$i]
	    }
	}
	Line    { set descr "$X0 $Y0 $X1 $Y1" }
	Point   { set descr "$X0 $Y0" }
    }
    return $descr
}

proc powConv2NewDescr { shape descr } {
# Convert the shape specific description to standard 2n+1 parameters
#     rot x0 y0 x1 y1 (... xn yn for polygons) in canvas coords
    global regionParam powPlotParam
    
    set cnt 0
    foreach p $descr {
	incr cnt
	set p$cnt $p
    }

    set gn $regionParam(gn)
    foreach {x0 y0} [powGraphToCanvas $gn $p1 $p2] {}
    if {$cnt>2} {
	set dx [expr $p3*$powPlotParam(xmagstep,$gn)]
	if {$cnt>3} {
	    set dy [expr $p4*$powPlotParam(ymagstep,$gn)]
	    if {$cnt>4} {
		set rot $p5
	    }
	}
    }

    switch $shape {
	Box     { set descr "$rot $x0 $y0 [expr $x0+0.5*$dx] [expr $y0+0.5*$dy]" }
	Circle  { set descr "0 $x0 $y0 [expr $x0+$dx/1.41421356] \
                     [expr $y0+$p3*$powPlotParam(ymagstep,$gn)/1.41421356]" }
	Ellipse { set descr "$rot $x0 $y0 [expr $x0+$dx/1.41421356] \
		                     [expr $y0+$dy/1.41421356]" }
        Polygon {
	    set newDescr "0 $x0 $y0"
	    foreach {x y} [lrange $descr 2 end] {
		foreach {x y} [powGraphToCanvas $gn $x $y] {}
		lappend newDescr $x $y
	    }
	    set descr $newDescr
	}
	Line    { 
	    foreach {x1 y1} [powGraphToCanvas $gn $p3 $p4] {}
	    set descr "0 $x0 $y0 $x1 $y1"}
	Point   { set descr "0 $x0 $y0 $x0 $y0" }
    }
    return $descr
}

proc powConvDesc2Fctn { shape descr {format default} } {
    global regionParam currimg

# Convert description to desired coordinate system if necessary

    if {[powWCSexists $regionParam(gn)]} {
	set WCS 1
    } else {
	set WCS 0
    }
    if {$format == "default"} {set format $regionParam(coords)}
    if { $WCS && $format=="pixels" } {
	if { $shape=="Line" || $shape=="Polygon" || $shape=="Point" } {
	    set newD ""
	    foreach {x y} $descr {
		foreach {x y} [powGraphToPixel $currimg $x $y] {}
		lappend newD [expr $x+1] [expr $y+1]
	    }
	    set descr $newD
	} else {
	    foreach {x y} [lrange $descr 0 1] {}
	    foreach {x y} [powGraphToPixel $currimg $x $y] {}
	    set descr [lreplace $descr 0 1 [expr $x+1] [expr $y+1]]
	}
    } elseif { !$WCS && $format=="degrees" } {
	tk_messageBox -type ok -message \
		"Region coded in degrees, but graph lacks WCS information."
	return ""
    }

    set n [llength $descr]
    set txt "("
    set comma ""
    for {set i 0} {$i<$n} {incr i} {
	set nmbr [format "%.6g" [lindex $descr $i] ]
	set txt "$txt$comma $nmbr"
	set comma ","
    }
    set txt "$txt)"
    return $txt
}

proc powConvFctn2Desc { descr {format default} } {
    global regionParam currimg

    set sign $regionParam(currSign)
    set shape $regionParam(currShape)

    set leftParen [string first "(" $descr]
    set rightParen [string first ")" $descr]
    if {$leftParen==-1 || $rightParen==-1 || \
	    $leftParen>=[expr $rightParen-1]} {return ""}

    if {$leftParen>0} {
	set shp [string range $descr 0 [expr $leftParen-1] ] 
	set shp [string toupper $shp]
	if { [string first "-" $shp]!=-1 } {set sign "-"} else {set sign "+"}
	switch -regexp -- $shp {
	    BOX { set shape "Box"}
	    ELLIPSE { set shape "Ellipse"}
	    POLYGON { set shape "Polygon"}
	    CIRCLE {set shape "Circle"}
	    LINE {set shape "Line"}
	    POINT {set shape "Point"}
	}
    }
    
    switch $shape {
	Box {set nparam 5}
	Ellipse {set nparam 5}
	Circle {set nparam 3}
	Polygon {set nparam -2}
	Line {set nparam 4}
	Point {set nparam 2}
    }

    set nmbrs [string range $descr [expr $leftParen+1] [expr $rightParen-1]]
    set elem [split $nmbrs ", "]
    
    set descr ""
    set nelem 0
    foreach item $elem {
	if {$item == ""} {continue}
	lappend descr $item
	incr nelem
    }
    if {$nparam==-2} {
	set nparam [expr 2*int($nelem/2)]
    }
    if {$nelem!=$nparam} {
	tk_messageBox -message "Wrong number of parameters for $shape!" \
		-type ok
	return ""
    }

# Convert description to the native coordinate system if necessary

    if {[powWCSexists $regionParam(gn)]} {
	set WCS 1
    } else {
	set WCS 0
    }
    if {$format == "default"} {set format $regionParam(coords)}
    if { $WCS && $format=="pixels" } {
	if { $shape=="Line" || $shape=="Polygon" || $shape=="Point" } {
	    set newD ""
	    foreach {x y} $descr {
		foreach {x y} [powPixelToGraph $currimg \
			                       [expr $x-1] [expr $y-1]] {}
		lappend newD $x $y
	    }
	    set descr $newD
	} else {
	    foreach {x y} [lrange $descr 0 1] {}
	    foreach {x y} [powPixelToGraph $currimg [expr $x-1] [expr $y-1]] {}
	    set descr [lreplace $descr 0 1 $x $y]
	}
    } elseif { !$WCS && $format=="degrees" } {
	tk_messageBox -type ok -message \
		"Region coded in degrees, but graph lacks WCS information."
	return ""
    }

    return [list $sign $shape $descr]
}


#########
#
#   Object coordinates for primitive shapes
#

# generate a list of points for a circle. every 12 degrees should do
    
set circle_points ""
set cnvt [expr -3.1415926535 / 180.0]
for {set i 0} {$i <= 360} {incr i 12} {
    lappend circle_points [expr cos($i * $cnvt)]
    lappend circle_points [expr sin($i * $cnvt)]
}


# return coords of unit circle, radius 1 at the origin

proc circle {} {
    global circle_points
    return $circle_points
}

# return coords of square (bounding box of unit circle)

proc square {} {
    return "1 1 1 -1 -1 -1 -1 1 1 1"
}

# rotate a polygon around x0 and y0, return new coordinates

proc poly_rotate {x0 y0 angle coords} {
   set st [expr sin(-3.1415926535 * $angle / 180.0)]
   set ct [expr cos(-3.1415926535 * $angle / 180.0)]
   set result ""
   foreach {x y} $coords {
       lappend result [expr $ct*$x - $st*$y + $x0*(1.0 - $ct) + $y0*$st]
       lappend result [expr $st*$x + $ct*$y + $y0*(1.0 - $ct) - $x0*$st]
   }
   return $result
}

proc ShapeToPoly { shape descr } {
    set rot [lindex $descr 0]
    set cnt 0
    foreach {x y} [lrange $descr 1 end] { 
	set x$cnt $x
	set y$cnt $y
	incr cnt
    }

    if {$cnt>1} {
	set dx [expr $x1-$x0]
	set dy [expr $y1-$y0]
    }

    switch $shape {
	Box     { set coords [BoxToPoly $x0 $y0 $dx $dy $rot] }
	Circle  { set coords [CircleToPoly $x0 $y0 $dx $dy] }
	Ellipse { set coords [EllipseToPoly $x0 $y0 [expr 1.41421356*$dx] \
		                                    [expr 1.41421356*$dy] $rot] }
        Polygon { set coords [PolyToPoly $descr] }
	Line    { set coords [list $x0 $y0 $x1 $y1 $x0 $y0] }
	Point   { set coords [PointToPoly $x0 $y0] }
    }
    return $coords
}

proc CircleToPoly { x0 y0 dx dy } {
    global powPlotParam regionParam

    set gn $regionParam(gn)
    set xRad [expr $dx/$powPlotParam(xmagstep,$gn)]
    set yRad [expr $dy/$powPlotParam(ymagstep,$gn)]
    set radius [expr sqrt($xRad*$xRad + $yRad*$yRad)]
    set xRad [expr $radius*$powPlotParam(xmagstep,$gn)]
    set yRad [expr $radius*$powPlotParam(ymagstep,$gn)]
    set points ""
    foreach {x y} [circle] {
	set x [expr $xRad*$x+$x0]
	set y [expr $yRad*$y+$y0]
	lappend points $x $y
    }
    return $points
}

proc EllipseToPoly { x0 y0 dx dy angle } {
    set points ""
    foreach {x y} [circle] {
	set x [expr $dx*$x+$x0]
	set y [expr $dy*$y+$y0]
	lappend points $x $y
    }
    if {[expr abs($angle)] < 1e-10} {return $points}
    return [poly_rotate $x0 $y0 $angle $points]
}

proc BoxToPoly { x0 y0 dx dy angle } {
    set points ""
    foreach {x y} [square] {
	set x [expr $dx*$x+$x0]
	set y [expr $dy*$y+$y0]
	lappend points $x $y
    }
    if {[expr abs($angle)] < 1e-10} {return $points}
    return [poly_rotate $x0 $y0 $angle $points]
}

proc PolyToPoly { descr } {
    lappend descr [lindex $descr 1] [lindex $descr 2]
    return [lrange $descr 1 end]
}

proc PointToPoly { x0 y0 } {
    global powPlotParam regionParam

    set gn $regionParam(gn)
    set halfxMag [expr 0.5*$powPlotParam(xmagstep,$gn)]
    set halfyMag [expr 0.5*$powPlotParam(ymagstep,$gn)]
    if {$halfxMag<1} {set halfxMag 1}
    if {$halfyMag<1} {set halfyMag 1}
    set x1 [expr $x0-$halfxMag]
    set y1 [expr $y0-$halfyMag]
    set x2 [expr $x0+$halfxMag]
    set y2 [expr $y0+$halfyMag]
    return "$x1 $y1 $x1 $y2 $x2 $y2 $x2 $y1 $x1 $y1"
}

#########
#
#   Handle region clipping
#

proc powConvPoly { P } {

    set PolyGrid ""
    set x [lindex $P 0]
    set y [lindex $P 1]
    for {set i 2} {$i<[llength $P]} {incr i 2} {
	set nextX [lindex $P $i]
	set nextY [lindex $P [expr $i+1]]
	set dx [expr ($nextX-$x)]
	set dy [expr ($nextY-$y)]
	if { $dx || $dy } {
	    lappend PolyGrid [list $x $y $dx $dy]
	}
	set x $nextX
	set y $nextY
    }
    return $PolyGrid
}

proc powClipPolys { P1 P2 } {
#
# Find the intersect region of two polygons.  P1 and P2 *must* be closed
# (ie, P[0]==P[last]) without any adjacent duplicate entries (P[i]=P[i+1])
# If a polygon is irregular (eg, U-shaped), this routine may return a list
# of lists of coordinates, mapping out the disjointed regions.

    set TINY 1e-10
    set ONEPLUS [expr 1+$TINY]
    set ONEMNUS [expr 1-$TINY]

    set Poly1 [powConvPoly $P1]
    set Poly2 [powConvPoly $P2]
    set nPoly1 [llength $Poly1]
    set nPoly2 [llength $Poly2]
    set Ipts ""

    if { !$nPoly1 || !$nPoly2 } {
	if {$nPoly1 \
		&& [powPtInRgn [lindex P2 0] [lindex P2 1] $Poly1]} {
	    return $P2
	} elseif { $nPoly2 \
		&& [powPtInRgn [lindex P1 0] [lindex P1 1] $Poly2]} {
	    return $P1
	}
	return ""
    }

# Find the Intersections of the two regions

    for {set i1 0} {$i1 < $nPoly1} {incr i1} {
	set Seg1 [lindex $Poly1 $i1]
	foreach {a_x a_y a_dx a_dy} $Seg1 {}

	for {set i2 0} {$i2 < $nPoly2} {incr i2} {
	    set Seg2 [lindex $Poly2 $i2]
	    foreach {b_x b_y b_dx b_dy} $Seg2 {}
	    
	    set a $i1
	    set b $i2
	    set num [expr $a_dx*($a_y-$b_y) - $a_dy*($a_x-$b_x)]
	    set den [expr $a_dx*$b_dy - $b_dx*$a_dy]
	    if {$den!=0 || ($den==0 && $num==0)} {
		if {$den==0} {
		    if {$b_dx} {
			set fb [expr ($a_x-$b_x)/$b_dx]
		    } else {
			set fb [expr ($a_y-$b_y)/$b_dy]
		    }
		    if {$fb>=0 && $fb<$ONEPLUS} {
			if {$fb>$ONEMNUS} {
			    incr b
			    if {$b==$nPoly2} {set b 0}
			} elseif { $fb>$TINY } {
			    incr b
			    set dx [expr $fb*$b_dx]
			    set dy [expr $fb*$b_dy]
			    set x  [expr $b_x+$dx]
			    set y  [expr $b_y+$dy]
			    set Seg2 [list $b_x $b_y $dx $dy]
			    set newSeg [list $x $y [expr $b_dx-$dx] \
				    [expr $b_dy-$dy] ]
			    set Poly2 [lreplace $Poly2 $i2 $i2 $Seg2 $newSeg]
			    incr nPoly2
			    foreach {b_dx b_dy} "$dx $dy" {}
			    set tmp ""
			    foreach j $Ipts {
				foreach {j1 j2} $j {}
				if {$j2>$i2} {incr j2}
				lappend tmp [list $j1 $j2]
			    }
			    set Ipts $tmp
			}
			set newI [list $a $b]
			if { [lsearch -exact $Ipts $newI]==-1 } {
			    lappend Ipts $newI
			}
			set b $i2
		    }
		    set fb 0
		} else {
		    set fb [expr $num/$den]
		}
		if { $fb>=0 && $fb<$ONEPLUS } {
		    if {$a_dx} {
			set fa [expr ($b_dx*$fb+$b_x-$a_x)/$a_dx]
		    } else {
			set fa [expr ($b_dy*$fb+$b_y-$a_y)/$a_dy]
		    }
		    if { $fa>=0 && $fa<$ONEPLUS } {
			if { $fa>$ONEMNUS } {
			    incr a
			    if {$a==$nPoly1} {set a 0}
			} elseif { $fa>$TINY } {
			    incr a
			    set dx [expr $fa*$a_dx]
			    set dy [expr $fa*$a_dy]
			    set x  [expr $a_x+$dx]
			    set y  [expr $a_y+$dy]
			    set Seg1 [list $a_x $a_y $dx $dy]
			    set newSeg [list $x $y [expr $a_dx-$dx] \
				    [expr $a_dy-$dy] ]
			    set Poly1 [lreplace $Poly1 $i1 $i1 $Seg1 $newSeg]
			    incr nPoly1
			    foreach {a_dx a_dy} "$dx $dy" {}
			    set tmp ""
			    foreach j $Ipts {
				foreach {j1 j2} $j {}
				if {$j1>$i1} {incr j1}
				lappend tmp [list $j1 $j2]
			    }
			    set Ipts $tmp
			}
			if { $fb>$ONEMNUS } {
			    incr b
			    if {$b==$nPoly2} {set b 0}
			} elseif { $fb>$TINY } {
			    incr b
			    set dx [expr $fb*$b_dx]
			    set dy [expr $fb*$b_dy]
			    set x  [expr $b_x+$dx]
			    set y  [expr $b_y+$dy]
			    set Seg2 [list $b_x $b_y $dx $dy]
			    set newSeg [list $x $y [expr $b_dx-$dx] \
				    [expr $b_dy-$dy] ]
			    set Poly2 [lreplace $Poly2 $i2 $i2 $Seg2 $newSeg]
			    incr nPoly2
			    foreach {b_dx b_dy} "$dx $dy" {}
			    set tmp ""
			    foreach j $Ipts {
				foreach {j1 j2} $j {}
				if {$j2>$i2} {incr j2}
				lappend tmp [list $j1 $j2]
			    }
			    set Ipts $tmp
			    incr i2
			}
			set newI [list $a $b]
			if { [lsearch -exact $Ipts $newI]==-1 } {
			    lappend Ipts $newI
			}
		    }
		}
	    }
	}
    }
    
    set Ipts [lsort -command {powSortIntSects 0} $Ipts]
    set Jpts [lsort -command {powSortIntSects 1} $Ipts]
    set Npts [llength $Ipts]

#    powClipDump $Ipts $Poly1 $Poly2

# If there are no intersections or only 1, return:
#      P1 if P1 is inside P2
#      P2 if P1 encloses P2
#      empty if there is no overlap

# If there is only one intersection, make sure first point isn't lying
# right on the other Polygon.  If it is, use second point for test.

    set state [powPtInRgn [lindex $P1 0] [lindex $P1 1] $Poly2]
    if { $Npts==1 } {
	set i1 [lindex [lindex $Ipts 0] 0]
	set za [lindex [lindex $Ipts 0] 2]
	if { $i1==0 && $za==0 } {
	    set state [powPtInRgn [lindex $P1 2] [lindex $P2 3] $Poly2]
	}
	set Npts 0
    }

    if { ! $Npts } {
	if {$state} { return [list $P1] }
	set state [powPtInRgn [lindex $P2 0] [lindex $P2 1] $Poly1]
	if {$state} { return [list $P2] }
	return ""
    }

# Do loop over all the intersections and make sure they all make it
# into the clipped region... this allows for disjointed clip regions

  set Mclips ""
  for {set IntSects 0} {$IntSects<$Npts} {incr IntSects} {
    if { ![info exists Iused($IntSects)] } {

        set I $IntSects
	set J [lsearch -exact $Jpts [lindex $Ipts $I] ]
	set Idir 1
	set Jdir 1
	set endPt $I
	set errFlag 0

	foreach {ia1 ib1} [lindex $Ipts $I] {}
	set pt [lindex $Poly1 $ia1]
	foreach {x y dx dy} $pt {}
	set clipped "$x $y"
	set Iused($I) 1

    while { 1 } {

	foreach {ia1 ib1} [lindex $Ipts $I] {}
	foreach {x1 y1 dx1 dy1} [lindex $Poly1 $ia1] {}
	foreach {x2 y2 dx2 dy2} [lindex_wrap $Poly1 [expr $ia1-1] ] {}
	set posState [powPtInRgn [expr $x1+0.5*$dx1] [expr $y1+0.5*$dy1] $Poly2]
	set negState [powPtInRgn [expr $x2+0.5*$dx2] [expr $y2+0.5*$dy2] $Poly2]

	set doP1 1
	if { $posState && !$negState } {
	    set Idir 1
	} elseif { !$posState && $negState } {
	    set Idir -1
	} elseif { $posState } {
	    set prevI [lsearch -exact $Ipts [lindex_wrap $Jpts [expr $J-$Jdir]]]
	    set Idir [expr $I-$prevI]
	    if { $Idir<-1 } {set Idir  1}
	    if { $Idir> 1 } {set Idir -1}
	} else {
#                  Both directions outside region... 
#    If this wasn't first point tested, return to tracing other region.
#    Otherwise, it was a single point intersection, so just finish.
	    set doP1 0
	    if { $I!=$endPt } {
		set ia2 [lindex [lindex_wrap $Ipts [expr $I+1] ] 0]
		if { $ia1==$ia2 } {
		    incr I
		} else { incr I -1 }
		set J [lsearch -exact $Jpts [lindex_wrap $Ipts $I] ]
	    }
	}

	while { $doP1 } {

	    foreach {ia1 ib1} [lindex $Ipts $I] {}
	    incr I $Idir
	    if {$I==$Npts} {set I 0} elseif {$I<0} {set I [expr $Npts-1]}
	    
# Copy over all polygon1 segments inside polygon2 (ia1->ia2)

	    foreach {ia2 ib2} [lindex $Ipts $I] {}
	    while {$ia1!=$ia2} {
		incr ia1 $Idir
		if {$ia1==$nPoly1} {set ia1 0} \
			elseif {$ia1<0} {set ia1 [expr $nPoly1-1]}
		set pt [lindex $Poly1 $ia1]
		lappend clipped [lindex $pt 0]
		lappend clipped [lindex $pt 1]
	    }
	    set J [lsearch -exact $Jpts [lindex $Ipts $I] ]
	    set Iused($I) 1
	    if {$I==$endPt} {break}

	    if { [llength $clipped] > [expr 3*($nPoly1+$nPoly2)]} {
		set errFlag 1
		tk_messageBox -message "Got lost clipping regions!  Region is too complex."
		break
	    }
# Check whether we actually need to switch to P2... test whether the middle
# point of the next segment is still inside P2

	    if {$Idir==1} {
		foreach {x y dx dy} [lindex $Poly1 $ia2] {}
	    } else {
		foreach {x y dx dy} [lindex_wrap $Poly1 [expr $ia2-1]] {}
	    }
	    if { ![powPtInRgn [expr $x+0.5*$dx] [expr $y+0.5*$dy] $Poly2] } {
		break
	    }
	}
	if {$I==$endPt} {break}
	if {$errFlag} {break}

# Follow polygon2 around until re-intersect polygon1
# First, need to know what direction to go!

	foreach {ja1 jb1} [lindex $Jpts $J] {}
	foreach {x1 y1 dx1 dy1} [lindex $Poly2 $jb1] {}
	foreach {x2 y2 dx2 dy2} [lindex_wrap $Poly2 [expr $jb1-1] ] {}
	set posState [powPtInRgn [expr $x1+0.5*$dx1] [expr $y1+0.5*$dy1] $Poly1]
	set negState [powPtInRgn [expr $x2+0.5*$dx2] [expr $y2+0.5*$dy2] $Poly1]

	set doP2 1
	if { $posState && !$negState } {
	    set Jdir 1
	} elseif { !$posState && $negState } {
	    set Jdir -1
	} elseif { $posState } {
	    set prevJ [lsearch -exact $Jpts [lindex_wrap $Ipts [expr $I-$Idir]]]
	    set Jdir [expr $J-$prevJ]
	    if { $Jdir<-1 } {set Jdir  1}
	    if { $Jdir> 1 } {set Jdir -1}
	} else {
	    set jb2 [lindex [lindex_wrap $Jpts [expr $J+1] ] 1]
	    if { $jb1==$jb2 } {
		incr J
	    } else { incr J -1 }
	    set I [lsearch -exact $Ipts [lindex_wrap $Jpts $J] ]
	    set doP2 0
	}

	while { $doP2 } {

	    foreach {ja1 jb1} [lindex $Jpts $J] {}
	    incr J $Jdir
	    if {$J==$Npts} {set J 0} elseif {$J<0} {set J [expr $Npts-1]}

# Copy over all polygon2 segments inside polygon1 (jb1->jb2)

	    foreach {ja2 jb2} [lindex $Jpts $J] {}
	    while {$jb1!=$jb2} {
		incr jb1 $Jdir
		if {$jb1==$nPoly2} {set jb1 0} \
			elseif {$jb1<0} {set jb1 [expr $nPoly2-1]}
		set pt [lindex $Poly2 $jb1]
		lappend clipped [lindex $pt 0]
		lappend clipped [lindex $pt 1]
	    }
	    set I [lsearch -exact $Ipts [lindex $Jpts $J] ]
	    set Iused($I) 1
	    if {$I==$endPt} {break}

# Check whether we actually need to switch to P1... test whether the middle
# point of the next segment is still inside P1

	    if { [llength $clipped] > [expr 3*($nPoly1+$nPoly2)]} {
		set errFlag 1
		tk_messageBox -message "Got lost clipping regions!  Region is too complex."
		break
	    }

	    if {$Jdir==1} {
		foreach {x y dx dy} [lindex $Poly2 $jb2] {}
	    } else {
		foreach {x y dx dy} [lindex_wrap $Poly2 [expr $jb2-1]] {}
	    }
	    if { ![powPtInRgn [expr $x+0.5*$dx] [expr $y+0.5*$dy] $Poly1] } {
		break
	    }
	}
	if {$I==$endPt} {break}
	if {$errFlag} {break}

    }
    lappend Mclips $clipped
}
}
return $Mclips
}

proc powClipDump { Ints P1 P2 } {

    puts "\nPolygon1"
    set i 0
    foreach is $P1 {
 	puts "[eval [concat format \"%2d  %9.3f  %9.3f  %9.3f  %9.3f\" $i $is]]"
	incr i
    }
    puts "Polygon2"
    set i 0
    foreach is $P2 {
 	puts "[eval [concat format \"%2d  %9.3f  %9.3f  %9.3f  %9.3f\" $i $is]]"
	incr i
    }
    puts "Intersections: "
    set i 0
    foreach is $Ints {
 	puts "[eval [concat format \"%2d  %2d  %2d\" $i $is]]"
	incr i
    }
}

proc lindex_wrap { L i } {
    set n [llength $L]
    while { $i < 0 }   { incr i $n }
    while { $i >= $n } { incr i [expr -$n] }
    return [lindex $L $i]
}

proc powSortIntSects { elem a b } {
    set d [expr [lindex $a $elem] - [lindex $b $elem] ]
    if {$d==0} {
	set d [expr [lindex $a [expr 1-$elem] ] - [lindex $b [expr 1-$elem] ] ]
    }
    if { $d<0 } {return -1} elseif { $d>0 } {return 1} else {return 0}
}

proc powPtInRgn { x y Poly } {

    lappend Poly [lindex $Poly 0]
    set nPoly [llength $Poly]

    set next [lindex $Poly 0]
    foreach {n_x n_y n_dx n_dy} $next {}

    set flag 0
    for {set cnt 1} {$cnt < $nPoly} {incr cnt} {
	set nxt [list $n_x $n_y $n_dx $n_dy]
	foreach {b_x b_y b_dx b_dy} $nxt {}
	foreach {n_x n_y n_dx n_dy} [lindex $Poly $cnt] {}
	if { ($y>$b_y && $y>=$n_y) || ($y<$b_y && $y<=$n_y) \
		|| ($x>$b_x && $x>=$n_x) } {
	    continue
	}

# Check to see if x,y lies right on the segment

	if { $x>=$b_x || $x>$n_x } {
	    set dy [expr $y-$b_y]
	    if { [expr abs($b_dy)]<1e-10 } {
		if { [expr abs($dy)]<1e-10 } {
		    return 1
		} else {
		    continue
		}
	    }
	    set dx [expr $b_x + ($b_dx/$b_dy)*$dy - $x]
	    if { $dx < -1e-10 } {continue}
	    if { $dx < 1e-10 } {return 1}
	}

# There is an intersection! Make sure it isn't a V point.    

	if { $y!=$b_y } {
	    set flag [expr 1-$flag]
	} else {
	    set idx [expr $cnt-1]
	    while {1} {
		if {$idx} {incr idx -1} else {set idx [expr $nPoly-2]}
		set prevdy [lindex [lindex $Poly $idx] 3]
		if {$prevdy} {break}
	    }
	    if {$b_dy*$prevdy > 0} {
		set flag [expr 1-$flag]
	    }
	}
    }
    return $flag
}
